"""
This module contains custom exception types.
"""


class ShapeError(Exception):
    """Error that the shape of a variable is incorrect."""
    def __init__(self, message, actual=None, expected=None):
        if actual is not None and expected is not None:
            super().__init__('{} (actual {} != expected {})'.format(message, actual, expected))
        elif actual is not None and expected is None:
            super().__init__('{} (actual {})'.format(message, actual))
        elif actual is None and expected is not None:
            super().__init__('{} (expected {})'.format(message, expected))
        else:
            super().__init__(message)


class DtypeError(TypeError):
    """Error that the dtype of a variable is incorrect."""
    def __init__(self, message, actual=None, expected=None):
        if actual is not None and expected is not None:
            super().__init__('{} (actual {} != expected {})'.format(message, actual, expected))
        elif actual is not None and expected is None:
            super().__init__('{} (actual {})'.format(message, actual))
        elif actual is None and expected is not None:
            super().__init__('{} (expected {})'.format(message, expected))
        else:
            super().__init__(message)



"""
Convenience implementation of the standard Thompson Sampling algorithm.
"""
import numpy
import typing


def sample_batch(
    samples:numpy.ndarray, *, ids:typing.Sequence,
    batch_size:int, seed:typing.Optional[int]=None
) -> tuple:
    """
    Ranks all candidates by their samples.

    Parameters
    ----------
        samples : numpy.ndarray
            (S, C) array of posterior samples (S) for each candidate (C)
        ids : numpy.ndarray, list, tuple
            (C,) candidate identifiers
        batch_size : int
            size of the next measurement batch (B)

    Returns
    -------
        chosen_candidates : tuple
            (B,) chosen candidate ids for the batch
    """
    n_samples, n_candidates = samples.shape
    assert len(ids) == n_candidates
    ids = numpy.atleast_1d(ids)
    chosen_candidates = []
    random = numpy.random.RandomState(seed)
    for i in range(batch_size):
        # for every sample in the batch, randomize the column order
        # to prevent always selecting lower-numbered candidates when >=2 samples are equal
        col_order = random.permutation(n_candidates)
        idx = random.randint(n_samples, size=n_candidates)
        selected_samples = samples[:, col_order][idx, numpy.arange(n_candidates)]
        best_candidate = ids[col_order][numpy.argmax(selected_samples)]
        chosen_candidates.append(best_candidate)
    random.seed(None)
    return tuple(chosen_candidates)


def get_probabilities(samples:numpy.ndarray, nit:int=100_000):
    """Get thompson sampling probabilities from posterior.

    Parameters
    ----------
        samples : numpy.ndarray
            (S, C) array of posterior samples (S) for each candidate (C)
        nit : int
            how many thompson draws samples to draw for the estimation

    Returns
    -------
        probabilities : numpy.ndarray
            (C,) probabilities that the candidates are sampled
    """
    n_samples, n_candidates = samples.shape

    frequencies = numpy.zeros(n_candidates)
    for _ in range(nit):
        idx = numpy.random.randint(n_samples, size=n_candidates)
        selected_samples = samples[idx, numpy.arange(n_candidates)]
        frequencies[numpy.argmax(selected_samples)] += 1
    return frequencies / frequencies.sum()



import numpy
import pytest

from . import utils


class TestOptimization:
    def test_multi_start_fmin(self):
        def fun(x):
            """ 2D function with the global minimum at [0, 0]. """
            x, y = x
            return numpy.cos(x + numpy.pi) + numpy.cos(y + numpy.pi) + (x / 5)**2 + (y / 8)**2
        assert fun([0, 0]) == -2.0
        x0s = numpy.random.uniform(-10, 10, size=(100, 2))
        x_best, y_best = utils.multi_start_fmin(fun, x0s)
        assert numpy.shape(x_best) == (2,)
        assert numpy.isscalar(y_best)
        numpy.testing.assert_array_almost_equal(x_best, [0, 0], decimal=4)
        assert numpy.allclose(y_best, -2)
        pass



"""
Implementation of random fourier features (RFF).

Adapted from other implementations:
+ [Hernández-Lobato, 2014](https://bitbucket.org/jmh233/codepesnips2014/src/master/sourceFiles/sampleMinimum.m)
+ [Cornell-MOE](https://github.com/wujian16/Cornell-MOE/blob/master/pes/PES/sample_minimum.py)
+ [Bradford, 2018](https://github.com/Eric-Bradford/TS-EMO/blob/master/TSEMO_V3.m#L495)
"""
import functools
import hashlib
import h5py
import numba
import numpy
import os
import scipy.linalg
import scipy.stats
import typing


from . exceptions import ShapeError, DtypeError


def _compute_inverse(A:numpy.ndarray) -> numpy.ndarray:
    """Compute the inverse of a matrix.

    Uses `scipy.linalg.cho_factor`, but falls back to `scipy.linalg.inv` on errors.
    """
    try:
        A_cholesky = scipy.linalg.cho_factor(A)
        A_inverse = scipy.linalg.cho_solve(A_cholesky, numpy.eye(A.shape[0]))
        return A_inverse
    except scipy.linalg.LinAlgError:
        A_inverse = scipy.linalg.inv(A)
        return A_inverse


def _allow_1d_inputs(method):
    """
    Decorates a method requiring 2D inputs such that 1D inputs are automatically
    expanded but the dimensionality of the return value is decremented again.
    """
    @functools.wraps(method)
    def wrapper(self, x):
        x2d = numpy.atleast_2d(x)
        if numpy.ndim(x) == 1:
            return method(self, x2d)[0]
        else:
            return method(self, x2d)
    return wrapper


class RffApproximation:
    """
    A function that approximates a sample from a GP by random fourier features.

    Actual computation functions are njitted static methods to speed up evaluation and optimization.
    """
    def __init__(
        self,
        sqrt_2_alpha_over_m:float,
        W:numpy.ndarray, B:numpy.ndarray, sample_of_theta:numpy.ndarray
    ):
        """Creates an RFF function object from function features.

        Parameters
        ----------
        sqrt_2_alpha_over_m : float
        W : numpy.ndarray
            (M, D) array
        B : numpy.ndarray
            (M, 1) array
        sample_of_theta : numpy.ndarray
            (M,) vector
        """
        self.sqrt_2_alpha_over_m = sqrt_2_alpha_over_m
        self.W = W
        self.B = B
        self.sample_of_theta = sample_of_theta
        super().__init__()

    @property
    def M(self) -> int:
        """Number of fourier features."""
        return self.W.shape[0]

    @property
    def D(self) -> int:
        """Number of input dimensions."""
        return self.W.shape[1]

    @property
    def uuid(self) -> str:
        """ A unique hash of the RFF function, created from its parameter arrays. """
        components = [
            hashlib.md5(self.sqrt_2_alpha_over_m).hexdigest(),
            hashlib.md5(self.W).hexdigest(),
            hashlib.md5(self.B).hexdigest(),
            hashlib.md5(self.sample_of_theta).hexdigest(),
        ]
        return hashlib.md5(','.join(components).encode('ascii')).hexdigest().encode('ascii')

    @_allow_1d_inputs
    def __call__(self, x:numpy.ndarray) -> typing.Union[numpy.ndarray, float]:
        """Evalues an RFF approximation function, specified by the function features.

        Parameters
        ----------
        x : numpy.ndarray
            a 2D array of coordinates (?, D)

        Returns
        -------
        approx_y : numpy.ndarray
            function evaluations (?,)
        """
        return RffApproximation._evaluate(x, self.sqrt_2_alpha_over_m, self.W, self.B, self.sample_of_theta)

    @_allow_1d_inputs
    def grad(self, x:numpy.ndarray) -> numpy.ndarray:
        """Evaluates the gradient of an RFF approximation function, specified by the function features.

        Parameters
        ----------
        x : numpy.ndarray
            a 2D array of coordinates (?, D)

        Returns
        -------
        dydx : numpy.ndarray
            evaluations of the gradient w.r.t. x (?, D)
        """
        return RffApproximation._grad(x, self.sqrt_2_alpha_over_m, self.W, self.B, self.sample_of_theta)

    @staticmethod
    @numba.njit
    def _evaluate(
        x:numpy.ndarray,
        sqrt_2_alpha_over_m:float, W:numpy.ndarray, B:numpy.ndarray, sample_of_theta:numpy.ndarray
    ) -> numpy.ndarray:
        M, D = W.shape
        N = x.shape[0]
        assert x.shape == (N, D)
        phi_x = sqrt_2_alpha_over_m * numpy.cos(numpy.dot(W, x.T) + B)
        assert phi_x.shape == (M, N)
        approx_y = numpy.dot(phi_x.T, sample_of_theta)
        assert approx_y.shape == (N,)
        return approx_y

    @staticmethod
    @numba.njit
    def _grad(x:numpy.ndarray, sqrt_2_alpha_over_m:float, W:numpy.ndarray, B:numpy.ndarray, sample_of_theta:numpy.ndarray) -> numpy.ndarray:
        M, D = W.shape
        N = x.shape[0]
        assert x.shape == (N, D)
        temp = numpy.sin(numpy.dot(W, x.T) + B)
        assert temp.shape == (M, N)
        # the following is a refactor of the Cornell-MOE implementation that supports numba.njit (about 6x faster)
        gradient = numpy.empty((N, D))
        for n in range(N):
            gradient_of_phi_x = -sqrt_2_alpha_over_m * temp[:,n] * W.T
            gradient[n] = numpy.dot(gradient_of_phi_x, sample_of_theta)
        return gradient


def sample_rff(
    lengthscales:numpy.ndarray,
    scaling:float,
    noise:float,
    kernel_nu:float,
    X:numpy.ndarray,
    Y:numpy.ndarray,
    M:int,
) -> typing.Tuple[
    typing.Callable[[numpy.ndarray], typing.Union[numpy.ndarray, float]],
    typing.Callable[[numpy.ndarray], typing.Union[numpy.ndarray, float]],
    typing.Callable[[numpy.ndarray], numpy.ndarray],
]:
    """
    Samples an approximate function from a GP with given kernel function and zero mean.

    It returns not only the approximate function, but also a njit-decorated version and its njit-decorated gradient.
    The njit-decorated callables have a compilation overhead on the first call!
    After that they are about 3.5x and 6x faster than non-njitted alternatives.

    Parameters
    ----------
    lengthscales : numpy.ndarray
        lengthscale values for each dimension (D,). It corresponds to...
            '1/l²' from [Hernández-Lobato, 2014] implementation
            'l' from Cornell-MOE implementation
            'ell' from [Bradford, 2018] implementation, line 501
    scaling : float
        the kernel standard deviation sigma. (As in k(x,x') = scaling**2 * ExpQuad()) It corresponds to...
            'sqrt(sigma)' from [Hernández-Lobato, 2014] implementation
            'sqrt(alpha)' from Cornell-MOE implementation
            'sqrt(sf2)' from [Bradford, 2018] implementation, line 502
    noise : float
        the observation noise from the likelihood function. It corresponds to...
            'sigma0' from [Hernández-Lobato, 2014] implementation
            'noise' from Cornell-MOE implementation
            'sn2' from [Bradford, 2018] implementation, line 503
    kernel_nu : float
        degree of freedom of the GP kernel, for example
            `numpy.inf` for squared exponential / RBF / exponential quadratic
            1/2 for Matern12
            3/2 for Matern32
    X : numpy.ndarray
        coordinates of observations (?, D)
    Y : numpy.ndarray
        observed values at coordinates (?,)
    M : int
        number of fourier features to use in the approximation
        according to [Mutny & Krause, 2018], the approximation error scales O(m^(1/2))

    Returns
    -------
    f_rff : callable
        may be called with (?, D) or (D,) coordinates, returning (?,) or float respectively
    f_rff_njit : callable
        jit-compiled version of the above
    g_rff_njit : callable
        jit-compiled gradient, takes (?, D) or (D,) coordinates, returning (?, D) or (?, D) respectively
    """
    # check inputs
    X = numpy.atleast_2d(X)
    Y = numpy.atleast_1d(Y)
    D = X.shape[1]
    N_obs = len(Y)
    lengthscales = numpy.atleast_1d(lengthscales)
    if not X.shape == (N_obs, D):
        raise ShapeError('Shapes of X and Y do not match.', expected='(?, D), (?,)', actual=f'{X.shape}, {Y.shape}')
    if not lengthscales.shape == (D,):
        raise ShapeError('Lengthscales and data dimensions do not match.', lengthscales.shape, (D,))
    if not numpy.ndim(scaling) == 0:
        raise ShapeError('Argument "scaling" must be a scalar.')
    if not numpy.ndim(noise) == 0:
        raise ShapeError('Argument "noise" must be a scalar.')
    if not isinstance(kernel_nu, (int, float)) or kernel_nu <= 0:
        raise ValueError('Argument "kernel_nu" must be a positive-definite scalar.')

    # see Bradford 2018, equation 26
    alpha = scaling**2
    # just to avoid re-computing it all the time:
    sqrt_2_alpha_over_m = numpy.sqrt(2*alpha/M)

    # construct function to sample p(w) (see [Bradford, 2018], equations 27 and 28)
    if numpy.isinf(kernel_nu):
        def p_w(size:tuple) -> numpy.ndarray:
            return numpy.random.normal(loc=0, scale=1 / lengthscales, size=size)
    else:
        def p_w(size:tuple) -> numpy.ndarray:
            return scipy.stats.t.rvs(loc=0, scale=1 / lengthscales, df=kernel_nu, size=size)

    # sample function features
    W = p_w(size=(M, D))
    assert W.shape == (M, D)
    B = numpy.random.uniform(0, 2*numpy.pi, size=(M, 1))
    assert B.shape == (M, 1)

    zeta = sqrt_2_alpha_over_m * numpy.cos(numpy.dot(W, X.T) + B)
    assert zeta.shape == (M, N_obs)

    A = numpy.divide(numpy.dot(zeta, zeta.T), noise) + numpy.eye(M)
    A_inverse = _compute_inverse(A)

    assert A_inverse.shape == (M, M)
    mean_of_post_theta = numpy.divide(numpy.dot(numpy.dot(A_inverse, zeta), Y), noise)
    assert mean_of_post_theta.shape == (M,)

    variance_of_post_theta = A_inverse
    assert variance_of_post_theta.shape == (M, M)

    sample_of_theta = numpy.random.multivariate_normal(mean_of_post_theta, variance_of_post_theta)
    assert sample_of_theta.shape == (M,)

    # create a function object
    rff = RffApproximation(sqrt_2_alpha_over_m, W, B, sample_of_theta)
    # call the function with test data to make sure that it works
    X_test = numpy.array([X[0], X[0]])
    assert numpy.shape(rff(X_test)) == (2,)
    assert numpy.shape(rff.grad(X_test)) == (2, D)
    return rff


def save_rffs(rffs:typing.Sequence[RffApproximation], fp:os.PathLike):
    """
    Saves a collection of RffApproximations to an HDF5 file.

    Parameters
    ----------
    rffs : list
        a collection of RffApproximation objects, each with the same D and M
    fp : os.PathLike
        filepath for the HDF5 file to create

    Returns
    -------
    fp : os.PathLike
        the filepath passed as the input, now pointing to an HDF5 file with datasets:
        (N, 1) sqrt_2_alpha_over_m
        (N, M, D) W
        (N, M, 1) B
        (N, M) sample_of_theta
        (N,) uuid
    """
    # check the inputs
    N = len(rffs)
    if N == 0:
        raise ValueError('The [rffs] collection is empty.')
    M = rffs[0].M
    D = rffs[0].D
    for rff in rffs:
        if not rff.M == M and rff.D == D:
            raise ValueError('All elements in the [rff] collection must have the same D and M.')
    with h5py.File(fp, 'w') as hfile:
        for name, shape, dtype in [
            ('sqrt_2_alpha_over_m', (N, 1), float),
            ('W', (N, M, D), float),
            ('B', (N, M, 1), float),
            ('sample_of_theta', (N, M), float),
            ('uuid', (N,), 'S32'),
        ]:
            hfile.create_dataset(name, shape, dtype=dtype, data=[
                getattr(rff, name)
                for rff in rffs
            ])
    return fp


def load_rffs(fp:os.PathLike) -> typing.Sequence[RffApproximation]:
    """
    Loads a collection of RffApproximations from an HDF5 file.

    Parameters
    ----------
    fp : os.PathLike
        filepath to an HDF5 file containing datasets for:
        (N, 1) sqrt_2_alpha_over_m
        (N, M, D) W
        (N, M, 1) B
        (N, M) sample_of_theta
        (N,) uuid

    Returns
    -------
    rffs : list
        loaded RffApproximation objects
    """
    rffs = None
    with h5py.File(fp, 'r') as hfile:
        N, M, D = hfile['W'].shape
        rffs = [None] * N
        for n in range(N):
            rffs[n] = RffApproximation(
                sqrt_2_alpha_over_m=hfile['sqrt_2_alpha_over_m'][n],
                W=hfile['W'][n],
                B=hfile['B'][n],
                sample_of_theta=hfile['sample_of_theta'][n],
            )
            # validate by comparing function hashes
            expected = hfile['uuid'][n]
            actual = rffs[n].uuid
            if actual != expected:
                raise Exception(f'The uuid of the function at position {n} did not match. ({actual} != {expected})')
    return rffs



import numpy
import scipy.optimize
import typing


def multi_start_fmin(fun:typing.Callable[[numpy.ndarray], float], x0s:numpy.ndarray) -> typing.Tuple[numpy.ndarray, float]:
    """
    Helper function to run fmin-optimization from many start points.

    Parameters
    ----------
    fun : callable
        the function to minimize
    x0s : numpy.ndarray
        (N_starts, D) array of initial guesses

    Returns
    -------
    x_best : numpy.ndarray
        (D,) coordinate of the found minimum
    y_best : float
        function value at the minimum
    """
    x_peaks = [
        scipy.optimize.fmin(fun, x0=x0, disp=False)
        for x0 in x0s
    ]
    y_peaks = [
        fun(x)
        for x in x_peaks
    ]
    ibest = numpy.argmin(y_peaks)
    return x_peaks[ibest], y_peaks[ibest]



from . exceptions import DtypeError, ShapeError
from . rff import sample_rff, save_rffs, load_rffs
from . thompson import sample_batch, get_probabilities
from . utils import multi_start_fmin

__version__ = '1.0.1'



import pytest

from . import exceptions


class TestExceptions:
    def test_dtype_error(self):
        with pytest.raises(exceptions.DtypeError) as exinfo:
            raise exceptions.DtypeError('Just the message.')
        assert 'Just' in exinfo.value.args[0]

        with pytest.raises(exceptions.DtypeError) as exinfo:
            raise exceptions.DtypeError('With types.', actual=str)
        assert 'str' in exinfo.value.args[0]

        with pytest.raises(exceptions.DtypeError) as exinfo:
            raise exceptions.DtypeError('With types.', expected=float)
        assert 'float' in exinfo.value.args[0]

        with pytest.raises(exceptions.DtypeError) as exinfo:
            raise exceptions.DtypeError('With types.', actual=int, expected=str)
        assert 'int' in exinfo.value.args[0] and 'str' in exinfo.value.args[0]
        pass

    def test_shape_error(self):
        with pytest.raises(exceptions.ShapeError) as exinfo:
            raise exceptions.ShapeError('Just the message.')
        assert 'Just' in exinfo.value.args[0]

        with pytest.raises(exceptions.ShapeError) as exinfo:
            raise exceptions.ShapeError('With shapes.', actual=(2,3))
        assert '(2, 3)' in exinfo.value.args[0]

        with pytest.raises(exceptions.ShapeError) as exinfo:
            raise exceptions.ShapeError('With shapes.', expected='(2,3) or (5,6)')
        assert '(5,6)' in exinfo.value.args[0]

        with pytest.raises(exceptions.ShapeError) as exinfo:
            raise exceptions.ShapeError('With shapes.', actual=(), expected='(5,4) or (?,?,6)')
        assert '(?,?,6)' in exinfo.value.args[0]
        pass



import numpy
import pytest

from . import thompson


class TestThompsonSampling:
    @pytest.mark.parametrize('batch_size', [1_000, 5_000])
    @pytest.mark.parametrize('seed', [None, 123, 349857])
    def test_sample_batch(self, batch_size, seed):
        S = 500
        C = 3
        ids = [f'C{c:02d}' for c in range(C)]
        samples = numpy.random.uniform(
            low=[0, 0, -1],
            high=[0.2, 1, 0],
            size=(S, C)
        )
        numpy.testing.assert_array_equal(samples.shape, (S, C))

        batch = thompson.sample_batch(
            samples=samples, ids=ids,
            batch_size=batch_size, seed=seed
        )
        assert len(batch) == batch_size
        # all samples for C02 are less then all others
        assert 'C02' not in batch
        assert batch.count('C00') < batch.count('C01')
        pass

    def test_no_bias_on_sample_collisions(self):
        samples = numpy.array([
            [2, 2, 2],
            [2, 2, 2],
        ])
        batch = thompson.sample_batch(samples, ids=('A', 'B', 'C'), batch_size=100, seed=1234)
        assert batch.count('A') != 100
        assert batch.count('C') != 0
        pass

    @pytest.mark.xfail(reason='Probabilities are currently computed by brute force and non-exact.')
    def test_get_probabilities_exact_on_identical(self):
        samples = numpy.array([
            [1, 2, 3, 4, 5],
            [5, 3, 4, 2, 1],
            [1, 3, 4, 2, 5]
        ]).T
        S, C = samples.shape
        assert S == 5
        assert C == 3

        probabilities = thompson.get_probabilities(samples)
        numpy.testing.assert_array_equal(probabilities, [1/C]*C)
        pass



import pathlib
import re
import setuptools

__packagename__ = 'pyrff'
ROOT = pathlib.Path(__file__).parent


def get_version():
    VERSIONFILE = pathlib.Path(ROOT, __packagename__, '__init__.py')
    initfile_lines = open(VERSIONFILE, 'rt').readlines()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(VSRE, line, re.M)
        if mo:
            return mo.group(1)
    raise Exception('Unable to find version string in %s.' % (VERSIONFILE,))


__version__ = get_version()


setuptools.setup(
    name=__packagename__,
    packages=setuptools.find_packages(),
    version=__version__,
    description='Implementation of random fourier feature (RFF) approximations and Thompson sampling.',
    license='AGPLv3',
    long_description=open(pathlib.Path(ROOT, 'README.md')).read(),
    long_description_content_type='text/markdown',
    url='https://github.com/michaelosthege/pyrff',
    author='Michael Osthege',
    author_email='michael.osthege@outlook.com',
    classifiers=[
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    install_requires=[
        open(pathlib.Path(ROOT, 'requirements.txt')).readlines()
    ],
    python_requires='>=3.6'
)



import h5py
import numpy
import pathlib
import pytest
import tempfile

from . import exceptions
from . import rff


class TestRffHelpers:
    def test_allow_1d_inputs(self):
        class TestClass:
            @rff._allow_1d_inputs
            def compute_function(self, x):
                return numpy.atleast_2d(x) + 5

        obj = TestClass()
        x_1d = numpy.array([1, 2, 3, 4])
        result_1d = obj.compute_function(x_1d)
        assert result_1d.shape == (4,)
        numpy.testing.assert_array_equal(result_1d, x_1d + 5)

        x_2d = numpy.array([
            [2, 3],
            [4, 5]
        ])
        result_2d = obj.compute_function(x_2d)
        assert result_2d.shape == (2, 2)
        numpy.testing.assert_array_equal(result_2d, x_2d + 5)
        pass


class TestSampleRFF:
    @pytest.mark.parametrize('kwarg,override,error_cls,message_fragment', [
        ('X', numpy.random.uniform(size=(47, 8)), exceptions.ShapeError, 'X and Y'),
        ('Y', numpy.random.normal(size=(44,)), exceptions.ShapeError, 'X and Y'),
        ('lengthscales', [1,2,3], exceptions.ShapeError, 'Lengthscales'),
        ('scaling', [12, 7], exceptions.ShapeError, 'scaling'),
        ('noise', [1,2,3], exceptions.ShapeError, 'noise'),
        ('kernel_nu', 'squared_exponential', ValueError, 'kernel_nu'),
        ('kernel_nu', -1, ValueError, 'kernel_nu'),
        ('kernel_nu', 0, ValueError, 'kernel_nu'),
    ])
    def test_sample_rff_inputchecks(self, kwarg, override, error_cls, message_fragment):
        D = 8
        kwargs = dict(
            lengthscales=numpy.random.uniform(size=(D,)),
            scaling=1.2,
            noise=0.001,
            kernel_nu=3/2,
            X=numpy.random.uniform(0, 1, size=(34, D)),
            Y=numpy.random.normal(size=(34,)),
            M=200
        )
        numpy.random.seed(1234)
        rff.sample_rff(**kwargs)

        kwargs[kwarg] = override
        with pytest.raises(error_cls) as exinfo:
            rff.sample_rff(**kwargs)
        assert message_fragment in exinfo.value.args[0]
        pass

    @pytest.mark.parametrize('M', [100, 200, 300])
    def test_rff_1d(self, M):
        X = numpy.array([[0, 0.75, -0.75, -0.375, 1.125, 0.375, -1.125]]).T
        Y = numpy.array([
            0.17817326, -0.14260799, 0.13385016,
            0.44349725, -0.11945854, -0.28416883, -0.06767212
        ])
        mp = {
            'ls': numpy.array([0.33805151]),
            'scaling': numpy.array(0.87595652),
            'sigma': numpy.array(0.00041511)
        }

        numpy.random.seed(1234)
        fun = rff.sample_rff(
            lengthscales=mp['ls'],
            scaling=mp['scaling'],
            noise=mp['sigma'],
            kernel_nu=numpy.inf,
            X=X, Y=Y, M=M
        )
        assert fun.D == 1
        assert fun.M == M

        # run checks with randomly sampled coordinates
        N = 500
        D = X.shape[1]
        X = numpy.random.uniform(low=-2, high=2, size=(N, D))

        # check the output shapes
        assert fun(X).shape == (N,)
        assert fun.grad(X).shape == (N, D)

        # check correctness of gradient by manual differentiation
        epsilon = 0.000001
        grad_diff = numpy.zeros((N, D))
        for d in range(D):
            x_plus, x_minus = X.copy(), X.copy()
            x_plus[:, d] += epsilon
            x_minus[:, d] -= epsilon
            grad_diff[:, d] = (fun(x_plus) - fun(x_minus)) / (2*epsilon)
        numpy.testing.assert_allclose(fun.grad(X), grad_diff, atol=1e-6)
        pass

    @pytest.mark.parametrize('M', [100, 200, 300])
    def test_rff_2d(self, M):
        X = numpy.array([
            [2.91766097, 2.24035167],
            [0.05847315, -1.36898571],
            [-0.97848763, -1.69827441],
            [-1.34113714, -0.94010644],
            [2.17295361, -2.05980198],
            [-2.15467653, 1.54248168],
            [1.41794951, -0.86602145],
            [-0.9534419, 1.0008183],
            [-1.69739617, 0.36856189],
            [-2.25492733, -1.08158109],
            [2.71928324, -2.17585926],
            [0.41647858, 2.8539929],
            [0.02020237, 1.00598524],
            [-2.79485101, -0.2632838],
            [-2.06489182, -0.1437062],
            [-1.98178537, 2.37755002],
            [-0.75963745, -0.72184242],
            [2.14989953, 0.87636633],
            [0.50077019, 1.01010016],
            [-1.93324428, 2.0954881],
            [-0.34576451, 1.98880626],
            [1.5835243, 2.51814586],
            [-2.57655944, -2.0630075],
            [0.82136539, 0.33417412],
            [-1.84842772, -0.4460641],
            [0.08052125, -1.38373449],
            [0.59412011, -1.67895633],
            [-1.19482503, -2.7096331],
            [0.38592035, 2.61619323],
            [1.81816585, 1.1838309]
        ])
        Y = numpy.array([
            0.2684595, -3.12345661, -1.20876944, -2.0739321, -0.02077514,
            0.50824162, -2.01574938, 1.43596611, -0.81839906, -0.17958556,
            -0.02243076, -0.30501982, 3.87966439, -0.25128758, -0.99743626,
            -0.18646517, -4.7381341, 7.1723458, 5.83390888, -0.01031394,
            -0.10963341, -0.07243528, -0.10766074, -0.9695199, -1.10566236,
            -3.04485123, -1.64010435, 0.01366085, -0.23411911, 7.61113715
        ])
        mp = {
            'ls': numpy.array([3.86353836, 1.24398732]),
            'scaling': numpy.array(8.82195995),
            'sigma': numpy.array(0.9789097)
        }

        numpy.random.seed(1234)
        fun = rff.sample_rff(
            lengthscales=mp['ls'],
            scaling=mp['scaling'],
            noise=mp['sigma'],
            kernel_nu=numpy.inf,
            X=X, Y=Y, M=M
        )
        assert fun.D == 2
        assert fun.M == M

        # run checks with randomly sampled coordinates
        N = 500
        D = X.shape[1]
        X = numpy.random.uniform(low=-2, high=2, size=(N, D))

        # check the output shapes
        assert fun(X).shape == (N,)
        assert fun.grad(X).shape == (N, D)

        # check correctness of gradient by manual differentiation
        epsilon = 0.000001
        grad_diff = numpy.zeros((N, D))
        for d in range(D):
            x_plus, x_minus = X.copy(), X.copy()
            x_plus[:, d] += epsilon
            x_minus[:, d] -= epsilon
            grad_diff[:, d] = (fun(x_plus) - fun(x_minus)) / (2*epsilon)
        numpy.testing.assert_allclose(fun.grad(X), grad_diff, atol=1e-6)
        pass


class TestRffSaveLoad:
    def test_save_load_rffs(self):
        X = numpy.array([[0, 0.75, -0.75, -0.375, 1.125, 0.375, -1.125]]).T
        Y = numpy.array([
            0.17817326, -0.14260799, 0.13385016,
            0.44349725, -0.11945854, -0.28416883, -0.06767212
        ])
        mp = {
            'ls': numpy.array([0.33805151]),
            'scaling': numpy.array(0.87595652),
            'sigma': numpy.array(0.00041511)
        }

        with pytest.raises(ValueError) as execinfo:
            rff.save_rffs([], 'testfile.h5')
        assert 'empty' in execinfo.value.args[0]

        # sample a list of RFFs
        numpy.random.seed(1234)
        rffs = [
            rff.sample_rff(
                lengthscales=mp['ls'],
                scaling=mp['scaling'],
                noise=mp['sigma'],
                kernel_nu=numpy.inf,
                X=X, Y=Y, M=200
            )
            for i in range(20)
        ]

        with tempfile.TemporaryDirectory() as tmpdir:
            fp = pathlib.Path(tmpdir, 'testrffs.h5')
            rff.save_rffs(rffs, fp)

            with h5py.File(fp, 'r') as hfile:
                assert 'sqrt_2_alpha_over_m' in hfile
                assert 'W' in hfile
                assert 'B' in hfile
                assert 'sample_of_theta' in hfile
                assert 'uuid' in hfile

            rffs_loaded = rff.load_rffs(fp)
            assert len(rffs_loaded) == len(rffs)
            for r in range(len(rffs)):
                assert rffs_loaded[r].uuid == rffs[r].uuid

        pass
